<?php

namespace Drupal\elevenlabs_field;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Utility\Token;
use Drupal\elevenlabs_field\Exceptions\ElevenLabsFFmpegFailure;
use Drupal\field\Entity\FieldConfig;
use Drupal\file\FileStorageInterface;

/**
 * Takes care of concanating an file.
 */
class ElevenLabsFFmpegService {

  /**
   * Array with all instructions.
   */
  protected $instructions = [];

  /**
   * Total calculated time.
   */
  protected $totalTime = 0;

  /**
   * Temp file.
   */
  protected $tmpFile = '';

  /**
   * Field config.
   */
  protected $fieldSettings = [];

  /**
   * The entity type manager.
   */
  protected FileStorageInterface $fileStorage;

  /**
   * The file system.
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The token system.
   */
  protected Token $token;

  /**
   * Construtor of FFmpeg service.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityType
   *   The entity type manager interface.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system interface.
   * @param \Drupal\Core\Utility\Token $token
   *   The token system.
   */
  public function __construct(EntityTypeManagerInterface $entityType, FileSystemInterface $fileSystem, Token $token) {
    $this->fileStorage = $entityType->getStorage('file');
    $this->fileSystem = $fileSystem;
    $this->token = $token;
  }

  /**
   * Create an audio file from a field value.
   *
   * @param mixed $fieldValues
   *   The elevenlabs field values.
   * @param string $fileName
   *   The field name to save to.
   * @param string $entityType
   *   The entity type to save to.
   * @param string $bundle
   *   The bundle to save to.
   *
   * @return \Drupal\file\Entity\File
   *   The file entity.
   */
  public function concanateAudios($fieldValues, $fieldName, $entityType, $bundle) {
    $this->setInstructions($fieldValues);
    $this->runCommand();
    $this->ensureFieldSettings($fieldName, $entityType, $bundle);
    return $this->generateFileEntity();
  }

  /**
   * Generates the file entity.
   *
   * @return \Drupal\file\Entity\File
   *   The file entity.
   */
  protected function generateFileEntity() {
    $fileName = 'elevenlabs.mp3';
    $dir = $this->token->replace($this->fieldSettings['uri_scheme'] . '://' . rtrim($this->fieldSettings['file_directory'], '/'));
    $this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY);
    $destination = $dir . '/' . $fileName;
    $destination = $this->fileSystem->copy($this->tmpFile, $destination, FileSystemInterface::EXISTS_RENAME);
    /** @var \Drupal\file\Entity\File */
    $file = $this->fileStorage->create([
      'uri' => $destination,
    ]);
    $file->setPermanent();
    $file->save();
    // Cleanup.
    unlink($this->tmpFile);
    return $file;
  }

  /**
   * Ensures the field config.
   *
   * @param string $fileName
   *   The field name to save to.
   * @param string $entityType
   *   The entity type to save to.
   * @param string $bundle
   *   The bundle to save to.
   */
  private function ensureFieldSettings($fieldName, $entityType, $bundle) {
    $config = FieldConfig::loadByName($entityType, $bundle, $fieldName);
    $this->fieldSettings = $config->getSettings();
  }

  /**
   * Create instructions.
   *
   * @param mixed $fieldValues
   *   The elevenlabs field values.
   */
  private function setInstructions($fieldValues) {
    foreach ($fieldValues as $value) {
      if (!empty($value['target_id'])) {
        /** @var \Drupal\file\Entity\File */
        $file = $this->fileStorage->load($value['target_id']);
        $path = $this->fileSystem->realpath($file->getFileUri());
        $this->instructions[$value['target_id']] = [
          'filePath' => $path,
          'duration' => $this->getDuration($path),
          'offset' => $value['start_time'],
        ];
      }
    }
    $this->instructions = array_values($this->instructions);
  }

  /**
   * Run the command.
   */
  private function runCommand() {
    $command = $this->getCommand();
    exec($command . ' 2>&1', $response, $code);
    if ($code) {
      throw new ElevenLabsFFmpegFailure(implode(",", $response), $code);
    }
  }

  /**
   * Get the command.
   *
   * @return string
   *   The full command.
   */
  private function getCommand() {
    $this->tmpFile = $this->fileSystem->tempnam($this->fileSystem->getTempDirectory(), 'elevenlabs_') . '.mp3';
    $command = '/usr/bin/ffmpeg -y -nostdin';
    $midCommand = '';
    $endCommand = '';
    $i = 0;
    $lastDuration = 0;
    $totalOffset = 0;
    foreach ($this->instructions as $key => $value) {
      $command .= ' -i ' . $value['filePath'];
      if ($i) {
        $midCommand .= "[$i:a]adelay=" . $lastDuration . "|" . $lastDuration . "[a$i]; ";
      } else {
        $midCommand .= "[$i:a]atrim=0:" . ($this->totalTime*10) . "[a$i]; ";
      }
      $totalOffset += $this->instructions[($key + 1)]['offset'] ?? 0;
      $lastDuration = ($value['duration'] * 10) + $totalOffset;
      $endCommand .= "[a$i]";
      $i++;
    }

    $command .= ' -filter_complex "' . $midCommand . $endCommand . "amix=inputs=" . count($this->instructions) . ":duration=longest[avol]; [avol]volume=5.0\"";
    $command .= " -c:a libmp3lame $this->tmpFile";

    $i = 0;
    return $command;
  }

  /**
   * Get start time of a clip.
   *
   * @param string $path
   *   The file path.
   *
   * @return int
   *   The duration in ms.
   */
  private function getDuration($path) {
    $output = [];
    $command = '/usr/bin/ffmpeg -nostdin -i "' . $path . '" 2>&1 | grep "Duration"';
    exec("$command", $output);
    preg_match('/Duration\:(.*)start/i', $output[0], $match);
    $timeParts = explode(':', str_replace([',', '.'], '', $match[1]));
    $time = ($timeParts[0] * 360000) + ($timeParts[1] * 6000) + $timeParts[2];
    return $this->totalTime += $time;
  }

}
