<?php

namespace Drupal\elevenlabs_field\Plugin\Field\FieldType;

use Drupal\Component\Utility\Random;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TypedData\DataDefinition;

/**
 * Defines the 'elevenlabs' field type.
 *
 * @FieldType(
 *   id = "elevenlabs",
 *   label = @Translation("ElevenLabs Generation"),
 *   category = @Translation("AI"),
 *   default_widget = "elevenlabs",
 *   default_formatter = "elevenlabs_default"
 * )
 */
class ElevenLabsGenerationItem extends FieldItemBase {

  /**
   * {@inheritdoc}
   */
  public static function mainPropertyName() {
    return 'text';
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultFieldSettings()
  {
    return [
      'save_audio' => '',
      'concatenate' => FALSE,
    ] + parent::defaultFieldSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
    if ($this->text !== NULL) {
      return FALSE;
    }
    elseif ($this->speaker !== NULL) {
      return FALSE;
    }
    elseif ($this->playing_time !== NULL) {
      return FALSE;
    }
    elseif ($this->start_time !== NULL) {
      return FALSE;
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {

    $properties['text'] = DataDefinition::create('string')
      ->setLabel(t('Text'));
    $properties['speaker'] = DataDefinition::create('string')
      ->setLabel(t('Speaker'));
    $properties['playing_time'] = DataDefinition::create('integer')
      ->setLabel(t('Playing Time'));
    $properties['start_time'] = DataDefinition::create('integer')
      ->setLabel(t('Start Time'));
    $properties['history_item_id'] = DataDefinition::create('string')
      ->setLabel(t('History Item ID'));

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public function getConstraints() {
    $constraints = parent::getConstraints();

    /*$options['text']['NotBlank'] = [];

    $constraint_manager = \Drupal::typedDataManager()->getValidationConstraintManager();
    $constraints[] = $constraint_manager->create('ComplexData', $options);*/
    // @todo Add more constraints here.
    return $constraints;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {

    $columns = [
      'text' => [
        'type' => 'text',
        'size' => 'big',
      ],
      'speaker' => [
        'type' => 'varchar',
        'length' => 255,
      ],
      'playing_time' => [
        'type' => 'int',
        'size' => 'normal',
      ],
      'start_time' => [
        'type' => 'int',
        'size' => 'normal',
      ],
      'history_item_id' => [
        'type' => 'varchar',
        'length' => 255,
      ],
      'target_id' => [
        'description' => 'The ID of the file entity.',
        'type' => 'int',
        'unsigned' => TRUE,
      ],
    ];

    $schema = [
      'columns' => $columns,
      'indexes' => [
        'history_item_id' => ['history_item_id'],
      ],
      'foreign keys' => [
        'target_id' => [
          'table' => 'file_managed',
          'columns' => ['target_id' => 'fid'],
        ],
      ],
    ];

    return $schema;
  }

  /**
   * {@inheritdoc}
   */
  public static function generateSampleValue(FieldDefinitionInterface $field_definition) {

    $random = new Random();

    $values['text'] = $random->paragraphs(5);

    $values['speaker'] = '21m00Tcm4TlvDq8ikWAM';

    $values['history_item_id'] = '21m00Tcm4TlvDq8ikWAM';

    $values['playing_time'] = mt_rand(-1000, 1000);

    $values['start_time'] = mt_rand(-1000, 1000);

    $dirname = 'public://elevenlabs';
    \Drupal::service('file_system')->prepareDirectory($dirname, FileSystemInterface::CREATE_DIRECTORY);
    $destination = $dirname . '/' . $random->name(10, TRUE) . '.mp4';
    $data = $random->paragraphs(3);
    /** @var \Drupal\file\FileRepositoryInterface $file_repository */
    $file_repository = \Drupal::service('file.repository');
    $file = $file_repository->writeData($data, $destination, FileSystemInterface::EXISTS_ERROR);
    $values['target_id'] = $file->id();

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function fieldSettingsForm(array $form, FormStateInterface $form_state) {
    $element = [];
    $settings = $this->getSettings();

    $element['save_audio'] = [
      '#type' => 'select',
      '#title' => $this->t('Save in Audio field'),
      '#default_value' => $settings['save_audio'],
      '#description' => $this->t('This will be the field to generate the audio in. Otherwise the audio will just be available in ElevenLabs. Only fields that allow mp3 and the same amount of fields or more will work.'),
      '#options' => $this->getAudioFieldsForGeneration($form),
    ];

    $element['concatenate'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Concatenate File'),
      '#description' => $this->t('If FFmpeg is available on the server, you can concatenate a long discussion into one file.'),
      '#default_value' => $settings['concatenate'],
      '#attributes' => [
        'disabled' => $this->checkFFmpeg(),
      ],
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave() {
    // Empty text means no saving.
    if (empty($this->text)) {
      $this->setValue([]);
      return;
    }
    parent::presave();
    $settings = $this->getSettings();
    // Get original values to calculate if we need to generate new audio.
    $elevenLabsMap = [];
    foreach ($this->getEntity()->original->get($this->getFieldDefinition()->getName())->getValue() as $value) {
      // Only count started processes.
      if ($value['history_item_id']) {
        $elevenLabsMap[$value['history_item_id']] = [
          'text' => $value['text'],
          'speaker' => $value['speaker'],
          'target_id' => $value['target_id'],
        ];
      }
    }

    // If no elevenlabs id exists, generate audio.
    $generateAudio = FALSE;
    // Recopy to audio fields if something changed.
    if (empty($this->history_item_id)) {
      $generateAudio = TRUE;
    }
    // If the text or speaker changed, but not the file.
    else if (isset($elevenLabsMap[$this->history_item_id]) &&
      ($this->text != $elevenLabsMap[$this->history_item_id]['text'] ||
      $this->speaker != $elevenLabsMap[$this->history_item_id]['speaker'])) {
      $generateAudio = TRUE;
    }

    if ($generateAudio && !empty($this->text)) {
      $generator = \Drupal::service('elevenlabs_field.generator_service');
      $data = $generator->generateFile($this->text, $this->speaker);
      if (!empty($data['file'])) {
        // Remove old data if exists.
        if ($this->target_id) {
          $generator->removeFile($this->target_id);
        }
        $this->set('target_id', $data['file']->id());
        $this->set('history_item_id', $data['history_item_id']);
      }
    }

    if ($settings['save_audio']) {
      _elevenlabs_field_needs_copying($this->getEntity(), [
        'type' => $settings['concatenate'] ? 'concatenate' : 'save',
        'field' => $settings['save_audio'],
        'original_field' => $this->getFieldDefinition()->getName(),
      ]);
    }
  }

  private function checkFFmpeg() {
    $command = (PHP_OS == 'WINNT') ? 'where ffmpeg' : 'which ffmpeg';
    $result = shell_exec($command);
    return !$result;
  }

  private function getAudioFieldsForGeneration(array $form) {
    $options[''] = '-- No audio generation --';
    /** @var \Drupal\Core\Field\FieldDefinitionInterface */
    foreach ($form['#entity']->getFieldDefinitions() as $fieldDefinition) {
      if ($fieldDefinition->getType() == 'file' && ($fieldDefinition->getFieldStorageDefinition()->getCardinality() == -1 ||
        $fieldDefinition->getFieldStorageDefinition()->getCardinality() > $this->getFieldDefinition()->getFieldStorageDefinition()->getCardinality())) {
          $config = $fieldDefinition->getConfig($form['#entity']->bundle())->getSettings();
          $extensions = preg_split('/ (,| ) /', $config['file_extensions']);
          if (in_array('mp3', $extensions)) {
            $options[$fieldDefinition->getName()] = $fieldDefinition->getLabel();
          }
      }
    }
    return $options;
  }

}
